@@ -46,6 +46,11 @@
  * that match the interests of this WpObjectManager will immediately become
  * available to get through wp_object_manager_new_iterator() and the
  * WpObjectManager \c object-added signal will be emitted for all of them.
+ * However, note that if these objects need to be prepared (to activate some
+ * features on them), the emission of \c object-added will be delayed. To know
+ * when it is safe to access the initial set of objects, wait until the
+ * \c installed signal has been emitted. That signal is emitted asynchronously
+ * after all the initial objects have been prepared.
  *
  * \gproperties
  *
@@ -106,11 +111,11 @@
  * Flags: G_SIGNAL_RUN_FIRST
  * \endparblock
  *
- * \par object-changed
+ * \par objects-changed
  * \parblock
  * \code
  * void
- * object_changed_callback (WpObjectManager * self,
+ * objects_changed_callback (WpObjectManager * self,
  *                          gpointer user_data)
  * \endcode
  *
@@ -687,8 +692,9 @@ wp_object_manager_maybe_objects_changed (WpObjectManager * self)
     if (!self->idle_source) {
       g_autoptr (WpCore) core = g_weak_ref_get (&self->core);
       if (core) {
-        wp_core_idle_add (core, &self->idle_source,
-            (GSourceFunc) idle_emit_objects_changed, self, NULL);
+        wp_core_idle_add_closure (core, &self->idle_source,
+            g_cclosure_new_object (
+                G_CALLBACK (idle_emit_objects_changed), G_OBJECT (self)));
       }
     }
   }
@@ -748,7 +754,7 @@ on_proxy_ready (GObject * proxy, GAsyncResult * res, gpointer data)
   self->pending_objects--;
 
   if (!wp_object_activate_finish (WP_OBJECT (proxy), res, &error)) {
-    wp_message_object (self, "proxy activation failed: %s", error->message);
+    wp_debug_object (self, "proxy activation failed: %s", error->message);
   } else {
     wp_object_manager_add_object (self, proxy);
   }
@@ -1063,7 +1069,7 @@ expose_tmp_globals (WpCore *core, GAsyncResult *res, WpRegistry *self)
     WpGlobal *g = g_ptr_array_index (tmp_globals, i);
 
     /* if global was already removed, drop it */
-    if (g->flags == 0)
+    if (g->flags == 0 || g->id == SPA_ID_INVALID)
       continue;
 
     /* if old global is owned by proxy, remove it */
@@ -1093,7 +1099,7 @@ expose_tmp_globals (WpCore *core, GAsyncResult *res, WpRegistry *self)
       WpGlobal *g = g_ptr_array_index (tmp_globals, i);
 
       /* if global was already removed, drop it */
-      if (g->flags == 0)
+      if (g->flags == 0 || g->id == SPA_ID_INVALID)
         continue;
 
       wp_object_manager_add_global (om, g);
@@ -1306,6 +1312,7 @@ void
 wp_global_rm_flag (WpGlobal *global, guint rm_flag)
 {
   WpRegistry *reg = global->registry;
+  guint32 id = global->id;
 
   /* no flag to remove */
   if (!(global->flags & rm_flag))
@@ -1313,7 +1320,7 @@ wp_global_rm_flag (WpGlobal *global, guint rm_flag)
 
   wp_trace_boxed (WP_TYPE_GLOBAL, global,
       "remove global %u flag 0x%x [flags:0x%x, reg:%p]",
-      global->id, rm_flag, global->flags, reg);
+      id, rm_flag, global->flags, reg);
 
   /* global was owned by the proxy; by removing the flag, we clear out
      also the proxy pointer, which is presumably no longer valid and we
@@ -1345,15 +1352,26 @@ wp_global_rm_flag (WpGlobal *global, guint rm_flag)
       /* remove FEATURE_BOUND to destroy the underlying pw_proxy */
       wp_object_deactivate (WP_OBJECT (proxy), WP_PROXY_FEATURE_BOUND);
 
+      /* stop all in-progress activations */
+      wp_object_abort_activation (WP_OBJECT (proxy), "PipeWire proxy removed");
+
       /* if the proxy is not owning the global, unref it */
       if (global->flags == 0)
         g_object_unref (proxy);
     }
+
+    /* It's possible to receive consecutive {add, remove, add} events for the
+     * same id. Since the WpGlobal might not be destroyed immediately below,
+     * (e.g. it's in tmp_globals list), we must invalidate the id now, so that
+     * this WpGlobal is not used in reference to objects added later.
+     */
+    global->id = SPA_ID_INVALID;
+    wp_properties_setf (global->properties, PW_KEY_OBJECT_ID, NULL);
   }
 
   /* drop the registry's ref on global when it does not appear on the registry anymore */
   if (!(global->flags & WP_GLOBAL_FLAG_APPEARS_ON_REGISTRY) && reg) {
-    g_clear_pointer (&g_ptr_array_index (reg->globals, global->id), wp_global_unref);
+    g_clear_pointer (&g_ptr_array_index (reg->globals, id), wp_global_unref);
   }
 }
 
